package # hide from pause?
 My::Test::Project;

=head1 NAME

My::Test::Project - can do stuff

=head1 VERSION

0.00

=cut

use strict;
use warnings;
use POSIX;
require Symbol;
use base qw/Tie::Hash/;

=head1 AUTHOR

See L<App::Mypp>.

=cut

1;
